/**************************************************************************
 * I2C Bus Scanner
 *  Copyright (c) 2012  Vanderhaegen Bart 
 *						bvsystems.be
 *
 *  Target PIC: 16F628A
 *
 * This device can scan all 127 hardware addresses of an I2C bus. It
 * will connect the I2C hardware addresses from all chips that respond
 * to their address. Afterwards, the results are shown on 2 7-segment
 * displays.
 *
 *************************************************************************/

#include <pic.h>;

// define I2C communications
#define SCL TRISB1					// serial clock (requires pull-up)
#define SDA TRISB2					// serial data(requires pull-up
#define SCL_IN	RB1					//
#define SDA_IN	RB2					//

#define I2C_ADR_BUFFER_SIZE	30  	// number of addresses to store

// define shift registers
#define SERDATA RB3
#define SHIFTCLK	RB5
#define DISABLE_LCD RB6

// define buttons
#define BUTTON RB0
#define JMP_SHIFT_RESULT RB7

// Define hex characters for display
const char hexData[]= 
	{0x3F,0x06,0x5B,0x4F,0x66,0x6D,0x7D,0x07,
	 0x7F,0x6F,0x77,0x7C,0x39,0x5E,0x79,0x71};

// Define symbols
#define LCD_P	0x73
#define LCD_A	0x77
#define LCD_B	0x7C
#define LCD_u	0x1C
#define LCD_min	0x40
#define LCD_equ	0x48
#define LCD_clr	0x00
#define LCD_C	0x39
#define LCD_T	0x78
#define LCD_E	0x79

// PIC Timebase (internal oscillator)
#define _XTAL_FREQ 4000000

// define boolean values
#define true	1
#define false	0

// define global variables. This variable is used across the
// program. (it is set by the interrupt handler).
bit buttonPushed = 0;
bit busy=0;

// Function prototypes
void doSelfTest(void);
void waitForUserPush(void);
void waitForEndConfirmation(void);
void runScanner(char* addresses,char* units,char* dozens,char* results);
void showNumberOfResults(char units,char dozens);
void showResults(char * addresses,char results);
void delay(void);
void i2cStart(void);
void i2cStop(void);
bit waitForAck(void);
void clock(void);
void sendData(char Data);
void giveI2Ccommand(char address,char data);
void delay_ms(int ms);
void sendAck(void);
void wait500ms(void);
void wait1500ms(void);
void showValues(char digit1,char digit2);
void sendValue(char digit);
void checkFreeI2Cbus(void);





/**************************************************************************
 Interrupt handler
   (with some software anti-bounce)
 *************************************************************************/
void interrupt irq() {
	if (INTF) {
		buttonPushed=1;
		__delay_ms(50);	// I know...
		while(!BUTTON);
		INTF=0;
	}
}





/**************************************************************************
 * I2C bus scanner - Main Program
 *
 *************************************************************************/
void main(void) {

	// initialisation
	TRISB=0x10000001;
	PORTB=0x00;
	DISABLE_LCD=1;
	TRISA=0x00;
	PORTA=0x00;

	// setup I2C values
	SCL = 1;
	SDA = 1;
	SCL_IN = 0;
	SDA_IN = 0;

	// setup interrupt system
	INTEDG=0;
	INTE=1;	
	INTF=0;
	ei();

	// perform an LCD selftest
	doSelfTest();

	while(true) {
		// Step 1: Wait until the user presses the Pu button
		waitForUserPush();

		// Step 2: define the array with the scan results	
		char addresses[I2C_ADR_BUFFER_SIZE];
		char results=0x00;
		char units=0;
		char dozens=0;
		buttonPushed=0;

		// Step 3: start a full I2C system scan
		runScanner(addresses,&units,&dozens,&results);

		// Step 4: show the results (if no interruption occured)
		if (!buttonPushed) {
			// Step 4: Show the number of results
			showNumberOfResults(units,dozens);		

			// Step 5: 
			buttonPushed=0;
			showResults(addresses,results);
		}

		// Step 5: show the ending of the program
		waitForEndConfirmation();
	}
}





/**************************************************************************
 * Program Functions
 * Optimized for the SN74AHC595 8-bit shift register
 *************************************************************************/
// This function waits for a release of the I2C bus
void checkFreeI2Cbus(void) {
	buttonPushed=0;
	SCL=1;
	SDA=1;

	busy=0;

	for (char i=0;i<100;i++) {
		busy=(!SCL_IN || !SDA_IN);
		__delay_ms(1);
	}

	if (busy) {
		showValues(LCD_u,LCD_B);
		while(!buttonPushed);
	}
}




// This function performs an auto test of the LCD
void doSelfTest(void) {
	// first a selftest is performed
	for (char cycle=0;cycle<2;cycle++) {
		for (char digit=0;digit<10;digit++) {
			if (cycle==0)
				showValues(hexData[digit],LCD_clr);
			else
				showValues(LCD_clr,hexData[digit]);

			__delay_ms(90);
		}
	}
}


// This function waits until the user has pushed OK
void waitForUserPush(void) {
	// Step 1: show Pu (Push) message: wait until pushed.
	buttonPushed=0;
	showValues(LCD_u,LCD_P);
	while(!buttonPushed);
}


// This function waits until the user has pushed OK
void waitForEndConfirmation(void) {
	// step 7: show the end message
	showValues(LCD_E,LCD_T);
	
	// step 8: show button
	buttonPushed=0;
	while(!buttonPushed);
}


// This function will run the actual scanner
void runScanner(char* addresses,char* units,char* dozens,char* results) {
	buttonPushed=0;
	char adressesToGo=9;

	checkFreeI2Cbus();

	if (!busy) {
		for (char adrToTest=0;adrToTest<128 && !buttonPushed;adrToTest++) {
			if (!buttonPushed) {
				i2cStart();
				sendData(adrToTest<<1);
			
				if (waitForAck()==1) {
					if (*results<I2C_ADR_BUFFER_SIZE) {
						(*addresses++)=adrToTest;
						*results=*results+1;
			
						*units=*units+1;
						if (*units==10) {
							*units=0;
							*dozens=*dozens+1;
						}
					}
				}
			
				i2cStop();
			
				// show progress
				if (!(adrToTest%15)) 
					adressesToGo--;
		
				if (adrToTest%4)
					showValues(hexData[adressesToGo],LCD_clr);
				else
					showValues(hexData[adressesToGo],LCD_min);	
			}
		}
	}
}



// showNumberOfResults shows the number of results
void showNumberOfResults(char units,char dozens) {
	while(!buttonPushed) {
		// Show "C="
		showValues(LCD_equ,LCD_C);
		wait500ms();
			
		// Show counter#
		showValues(hexData[units],hexData[dozens]);
		wait1500ms();
	
		// Show "Pu"
		showValues(LCD_u,LCD_P);
		wait500ms();
	}
}



// showResults shows the actual I2C addresses
void showResults(char * addresses,char results) {
	
	for (char resultScroller=0;resultScroller<results;resultScroller++) {
		char address=*addresses;

		// shift the address with 1 bit - if required
		if (!JMP_SHIFT_RESULT) {
			address<<=1;
		}

		// get the upper and the lower byte
		char lsb=address&0x0F;
		char msb=(address&0xF0)>>4;


		while (!buttonPushed) {
			// Show "A="
			showValues(LCD_equ,LCD_A);
			wait500ms();
		
			// show address
			showValues(hexData[lsb],hexData[msb]);
			wait1500ms();
		
			// Show "Pu"
			showValues(LCD_u,LCD_P);
			wait500ms();
		}
			
		buttonPushed=0;
		*addresses++;	// pointer increase
	}
}





/**************************************************************************
 * Shift register LCD controller
 * Optimized for the SN74AHC595 8-bit shift register
 **************************************************************************************/
// sendValue will send some data over the shift registers
void sendValue(char digit) {
	for (char j=0;j<8;j++) {
		if(digit & 0x80)
			SERDATA=1;
		else
			SERDATA=0;

		// generate a clockpulse for the shift register
		SHIFTCLK=1; SHIFTCLK=0;

		digit<<=1;
	}
}


// showValues is used to send the data to the displays
void showValues(char digit1,char digit2) {
	// disable the shift register's output for a while
	DISABLE_LCD=1;

	// send the character data
	sendValue(digit1);
	sendValue(digit2);

	// send an additional clock pulse
	SERDATA=0;
	SHIFTCLK=1; SHIFTCLK=0;	

	// enable the shift register's output again
	DISABLE_LCD=0;
}





/**************************************************************************
 * Delay functions with time-out
 * See http://www.bvsystems.be/pic_loop.php for more information.
 *************************************************************************/
// Wait 500 ms
void wait500ms() {
	unsigned int timeout=38462;
	while (buttonPushed==0 && (--timeout>0));
}


// Wait 1500 ms
void wait1500ms() {
	unsigned long timeout=55556;
	while (buttonPushed==0 && (--timeout>0));
}





/**************************************************************************
 * Implementation of the I2C protocol
 *************************************************************************/

	// initiate a new I2C command
	void giveI2Ccommand(char address,char data) {
		i2cStart();	 
		sendData(address);
		waitForAck();
		sendData(data);
		waitForAck();
		i2cStop();
	}


	// send 8 bits of data over the I2C bus
	// Assumption: SCL is already low
	void sendData(char Data) {
		unsigned char i;
	
		for (i=8;i>0;i--) {
			if(Data & 0x80) {
				SDA=1;
			}
			else {
				SDA_IN=0; SDA=0;
			}

			SCL=1;
			delay();
			Data<<=1;
			SCL_IN=0; SCL=0;
			delay();
		}
	}


	// send the I2C start sequence
	// SDA must go from high to low while SCL is high.
	void i2cStart() {
		SDA=1;		
		delay();
		SCL=1;		
		delay();	
		SDA_IN=0; SDA=0; 			
		delay();
		SCL_IN=0; SCL=0; 
		delay();
	}

	
	// send the I2C stop sequence
	// SDA must go from low to high while SCL is high
	void i2cStop() {
		SDA_IN=0; SDA=0;		
		delay();	
		SCL=1;		
		delay();		
		SDA=1;		
		delay();
	}



	// wait for an acknowledge signal
	// Receiver must make SDA low while SCL is high
	// The waitForAck times out after approximately 10 ms

	bit waitForAck() {
		SDA=1;			
		delay();		
		SCL=1;	
		unsigned int timeout=1923;
		while (SDA_IN && (--timeout > 0));
		SCL_IN=0; SCL=0;		
		delay();
		return (timeout>0);
	}



	// send the clock pulse
	// SCL must be made high for a short period of time
	void clock() {
		delay();	
		SCL=1;		
		delay();		
		SCL_IN=0; SCL=0;		
		delay();	
	}


	// send the acknowledge pulse required by the slave device
	void sendAck() {
		SDA_IN=0; 
		SDA=0;		
		delay();
		SCL=1;			
		delay();		
		SCL_IN=0; SCL=0;	
		delay();		
	}


	// delay execution
	void delay(void){
		__delay_us(100);
	}